#!/usr/bin/env python

"""
Copyright (C) 2005 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from qt import SIGNAL, QString, QStringList
from kdecore import KInstance, KLibFactory, KApplication, KURL
from kparts import KParts
from dcopexport import DCOPExObj
from dcop import DCOPRef
import khtml
import xml.dom
import new
DOMString = khtml.DOM.DOMString

# Save the Python built-in 'id' function.

objid = id

# Define a class which will be initialised when the plugin is loaded.

class FactoryClass(KLibFactory):

    instance = None

    def __init__(self, parent=None, name=None):

        KLibFactory.__init__(self, parent, name)

        if FactoryClass.instance is None:
            FactoryClass.instance = KInstance("DCOPAutomationPlugin")

    def createObject(self, parent=None, name=None, class_name="QObject", args=None):

        if args is None:
            args = []

        plugin = DocumentPlugin(parent, name)
        self.emit(SIGNAL("objectCreated(QObject)"), (plugin,))
        return plugin

# Define a plugin which exports certain aspects of the current document.

class DocumentPlugin(KParts.Plugin):

    def __init__(self, parent=None, name=None):
        KParts.Plugin.__init__(self, parent, name)
        self.automation = AutomationRoot(plugin=self)
        self.connect(self, SIGNAL("destroyed()"), self.stopAutomation)
        if parent.inherits("KHTMLPart"):
            self.connect(parent, SIGNAL("started(KIO::Job *)"), self.pauseAutomation)
            self.connect(parent, SIGNAL("completed()"), self.startAutomation)

    def stopAutomation(self):
        if self.automation is not None:
            self.automation.stopAutomation()
            del self.automation

    def startAutomation(self):
        self.automation.startAutomation()

    def pauseAutomation(self):
        self.automation.pauseAutomation()

class AutomationRoot(DCOPExObj):

    def __init__(self, id="AutomationRoot", plugin=None):
        DCOPExObj.__init__(self, "%s%s" % (id, objid(self)))
        self.plugin = plugin
        self.khtmlpart = plugin.parent()
        self.objects = {}
        self.ready = 0
        self.generation = 0

        # Top-level:

        self.addMethod("QString getDocumentSource()", self.getDocumentSource)
        self.addMethod("QString getURL()", self.getURL)
        self.addMethod("bool openURL(QString)", self.openURL)
        self.addMethod("bool closeURL()", self.closeURL)
        self.addMethod("QString baseURL()", self.baseURL)

        # Top-level status:

        self.addMethod("bool isReady()", self.isReady)
        self.addMethod("void setReady(bool)", self.setReady)
        self.addMethod("long getGeneration()", self.getGeneration)

        # Top-level (using the special result type - see below):

        self.addMethod("QStringList getDocument()", self.getDocument)

        # A union of all possible methods with the affected object as the first
        # argument. Most methods return QStringList with the first element being
        # a type indicator and the second element being either a QString value
        # or a QString reference to some kind of object. 

        # Node:

        self.addMethod("QStringList namespaceURI(QString)",                             self.string_method("namespaceURI"))
        self.addMethod("QStringList prefix(QString)",                                   self.string_method("prefix"))
        self.addMethod("QStringList localName(QString)",                                self.string_method("localName"))
        self.addMethod("QStringList nodeName(QString)",                                 self.string_method("nodeName"))
        self.addMethod("QStringList nodeValue(QString)",                                self.string_method("nodeValue"))
        self.addMethod("short nodeType(QString)",                                       self.other_method("nodeType"))
        self.addMethod("QStringList ownerDocument(QString)",                            self.object_method("ownerDocument", "Node"))
        self.addMethod("QStringList parentNode(QString)",                               self.object_method("parentNode", "Node"))
        self.addMethod("QStringList childNodes(QString)",                               self.object_method("childNodes", "NodeList"))
        self.addMethod("QStringList insertBefore(QString, QString, QString)",           self.action_method("insertBefore", "Node"))
        self.addMethod("QStringList replaceChild(QString, QString, QString)",           self.action_method("replaceChild", "Node"))
        self.addMethod("QStringList removeChild(QString, QString)",                     self.action_method("removeChild", "Node"))
        self.addMethod("QStringList appendChild(QString, QString)",                     self.action_method("appendChild", "Node"))
        self.addMethod("QStringList attributes(QString)",                               self.object_method("attributes", "NamedNodeMap"))
        self.addMethod("void normalize(QString)",                                       self.set_method("normalize"))
        self.addMethod("bool isNull(QString)",                                          self.other_method("isNull"))

        # Element:

        self.addMethod("QStringList tagName(QString)",                                  self.string_method("tagName"))
        self.addMethod("bool hasAttribute(QString, QString)",                           self.other_method("hasAttribute"))
        self.addMethod("bool hasAttributeNS(QString, QString, QString)",                self.other_method("hasAttributeNS"))
        self.addMethod("QStringList getAttribute(QString, QString)",                    self.string_method("getAttribute"))
        self.addMethod("QStringList getAttributeNode(QString, QString)",                self.object_method("getAttributeNode", "Node"))
        self.addMethod("QStringList getAttributeNS(QString, QString, QString)",         self.string_method("getAttributeNS"))
        self.addMethod("QStringList getAttributeNodeNS(QString, QString, QString)",     self.object_method("getAttributeNodeNS", "Node"))
        self.addMethod("void setAttribute(QString, QString, QString)",                  self.set_method("setAttribute"))
        self.addMethod("QStringList setAttributeNode(QString, QString)",                self.action_method("setAttributeNode", "Node"))
        self.addMethod("void setAttributeNS(QString, QString, QString, QString)",       self.set_method("setAttributeNS"))
        self.addMethod("QStringList setAttributeNodeNS(QString, QString)",              self.action_method("setAttributeNodeNS", "Node"))
        self.addMethod("void removeAttribute(QString, QString)",                        self.set_method("removeAttribute"))
        self.addMethod("QStringList removeAttributeNode(QString, QString)",             self.action_method("removeAttributeNode", "Node"))
        self.addMethod("void removeAttributeNS(QString, QString, QString)",             self.set_method("removeAttributeNS"))
        #self.addMethod("QStringList removeAttributeNodeNS(QString, QString)",          self.action_method("removeAttributeNodeNS", "Node"))
        self.addMethod("QStringList toHTML(QString)",                                   self.string_method("toHTML"))

        # Document:

        self.addMethod("QStringList createElement(QString, QString)",                   self.object_method("createElement", "Node"))
        self.addMethod("QStringList createElementNS(QString, QString, QString)",        self.object_method("createElementNS", "Node"))
        self.addMethod("QStringList createTextNode(QString, QString)",                  self.object_method("createTextNode", "Node"))
        self.addMethod("QStringList createAttribute(QString, QString)",                 self.object_method("createAttribute", "Node"))
        self.addMethod("QStringList createAttributeNS(QString, QString, QString)",      self.object_method("createAttributeNS", "Node"))
        self.addMethod("QStringList importNode(QString, QString, bool)",                self.importNode)
        self.addMethod("QStringList getElementById(QString, QString)",                  self.object_method("getElementById", "Node"))
        self.addMethod("QStringList getElementsByTagName(QString, QString)",            self.object_method("getElementsByTagName", "NodeList"))
        self.addMethod("QStringList getElementsByTagNameNS(QString, QString, QString)", self.object_method("getElementsByTagNameNS", "NodeList"))
        self.addMethod("QStringList toString(QString)",                                 self.string_method("toString"))

        # NodeList and NamedNodeMap:

        self.addMethod("long length(QString)",                                          self.other_method("length"))
        self.addMethod("QStringList item(QString, long)",                               self.item)

        # NamedNodeMap:

        self.addMethod("QStringList getNamedItem(QString, QString)",                    self.object_method("getNamedItem", "Node"))
        self.addMethod("QStringList getNamedItemNS(QString, QString, QString)",         self.object_method("getNamedItemNS", "Node"))
        self.addMethod("QStringList setNamedItem(QString, QString)",                    self.action_method("setNamedItem", "Node"))
        self.addMethod("QStringList setNamedItemNS(QString, QString)",                  self.action_method("setNamedItemNS", "Node"))
        self.addMethod("QStringList removeNamedItem(QString, QString)",                 self.object_method("removeNamedItem", "Node"))
        self.addMethod("QStringList removeNamedItemNS(QString, QString, QString)",      self.object_method("removeNamedItemNS", "Node"))

        # HTML-specific:

        self.addMethod("QStringList accessKey(QString)",                                self.string_method("accessKey"))
        self.addMethod("void setAccessKey(QString, QString)",                           self.set_method("setAccessKey"))
        self.addMethod("QStringList form(QString)",                                     self.object_method("form", "Node"))
        self.addMethod("QStringList name(QString)",                                     self.string_method("name"))
        self.addMethod("void setName(QString, QString)",                                self.set_method("setName"))
        self.addMethod("QStringList type(QString)",                                     self.string_method("type"))
        self.addMethod("void setType(QString, QString)",                                self.set_method("setType"))
        self.addMethod("QStringList value(QString)",                                    self.string_method("value"))
        self.addMethod("void setValue(QString, QString)",                               self.set_method("setValue"))
        self.addMethod("bool disabled(QString)",                                        self.other_method("disabled"))
        self.addMethod("void setDisabled(QString, bool)",                               self.flag_method("setDisabled"))
        self.addMethod("bool checked(QString)",                                         self.other_method("checked"))
        self.addMethod("void setChecked(QString, bool)",                                self.flag_method("setChecked"))
        self.addMethod("long maxLength(QString)",                                       self.other_method("maxLength"))
        self.addMethod("void setMaxLength(QString, long)",                              self.flag_method("setMaxLength"))
        self.addMethod("long size(QString)",                                            self.other_method("getSize"))
        self.addMethod("void setSize(QString, long)",                                   self.flag_method("setSize"))
        self.addMethod("QStringList accept(QString)",                                   self.string_method("accept"))
        self.addMethod("void setAccept(QString, QString)",                              self.set_method("setAccept"))
        self.addMethod("void blur(QString)",                                            self.set_method("blur"))
        self.addMethod("void focus(QString)",                                           self.set_method("focus"))
        self.addMethod("void select(QString)",                                          self.set_method("select"))
        self.addMethod("void click(QString)",                                           self.set_method("click"))
        self.addMethod("bool selected(QString)",                                        self.other_method("selected"))
        self.addMethod("void setSelected(QString, bool)",                               self.flag_method("setSelected"))
        self.addMethod("bool defaultSelected(QString)",                                 self.other_method("defaultSelected"))
        self.addMethod("void setDefaultSelected(QString, bool)",                        self.flag_method("setDefaultSelected"))
        self.addMethod("long selectedIndex(QString)",                                   self.other_method("selectedIndex"))
        self.addMethod("void setSelectedIndex(QString, long)",                          self.flag_method("setSelectedIndex"))
        self.addMethod("bool multiple(QString)",                                        self.other_method("multiple"))
        self.addMethod("void setMultiple(QString, bool)",                               self.flag_method("setMultiple"))
        self.addMethod("long length(QString)",                                          self.other_method("length"))
        self.addMethod("QStringList options(QString)",                                  self.object_method("options", "NodeList"))
        self.addMethod("long cols(QString)",                                            self.other_method("cols"))
        self.addMethod("void setCols(QString, long)",                                   self.flag_method("setCols"))
        self.addMethod("long rows(QString)",                                            self.other_method("rows"))
        self.addMethod("void setRows(QString, long)",                                   self.flag_method("setRows"))
        self.addMethod("QStringList defaultValue(QString)",                             self.string_method("defaultValue"))
        self.addMethod("void setDefaultValue(QString, QString)",                        self.set_method("setDefaultValue"))
        self.addMethod("QStringList acceptCharset(QString)",                            self.string_method("acceptCharset"))
        self.addMethod("void setAcceptCharset(QString, QString)",                       self.set_method("setAcceptCharset"))
        self.addMethod("QStringList action(QString)",                                   self.string_method("action"))
        self.addMethod("void setAction(QString, QString)",                              self.set_method("setAction"))
        self.addMethod("QStringList enctype(QString)",                                  self.string_method("enctype"))
        self.addMethod("void setEnctype(QString, QString)",                             self.set_method("setEnctype"))
        self.addMethod("QStringList method(QString)",                                   self.string_method("method"))
        self.addMethod("void setMethod(QString, QString)",                              self.set_method("setMethod"))
        self.addMethod("QStringList target(QString)",                                   self.string_method("target"))
        self.addMethod("void setTarget(QString, QString)",                              self.set_method("setTarget"))
        self.addMethod("void submit(QString)",                                          self.set_method("submit"))
        self.addMethod("void reset(QString)",                                           self.set_method("reset"))

    def stopAutomation(self):
        self.pauseAutomation()
        self.methods = {}
        self.plugin = None

    def startAutomation(self):
        self.objects = {}
        self.ready = 1
        self.generation += 1

    def pauseAutomation(self):
        self.objects = {}
        self.ready = 0

    def isReady(self):
        return self.ready

    def setReady(self, ready):
        self.ready = ready

    def getGeneration(self):
        return self.generation

    def getDocumentSource(self):
        if not self.ready:
            return QString.null

        return self.khtmlpart.document().toString().string()

    def getURL(self):
        if not self.ready:
            return QString.null

        return self.khtmlpart.toplevelURL().url()

    def openURL(self, url):
        if not self.ready:
            return 0

        # NOTE: The pausing of automation may not be necessary if the started
        # NOTE: signal works properly.

        #self.pauseAutomation()
        return self.khtmlpart.openURL(KURL(url))

    def closeURL(self):
        if not self.ready:
            return 0

        return self.khtmlpart.closeURL()

    def baseURL(self):
        if not self.ready:
            return QString.null

        return self.khtmlpart.baseURL().url()

    def getDocument(self):
        if not self.ready:
            return self._returnObject(None, "Node")

        return self._returnObject(self.khtmlpart.document(), "Node")

    # Marshalling methods.

    def _getObject(self, arg):
        id = str(arg)
        return self._convert(self.objects.get(id))

    def _returnObject(self, obj, objType):
        result = QStringList()
        result.append(objType)
        if obj is None:
            result.append(QString.null)
        elif objType == "string":
            result.append(obj) # QString, in fact.
        else:
            self.objects[str(objid(obj))] = obj
            result.append(QString(str(objid(obj))))
        return result

    # NOTE: Extracted from qtxmldom.

    def _convert(self, node):
        if not hasattr(node, "nodeType"):
            return node

        if node.nodeType() == xml.dom.Node.ELEMENT_NODE:

            # Promote underlying nodes to more capable HTML element objects.
            # NOTE: This does not check to see if such conversions are valid.

            localName = unicode(node.nodeName().string()).lower()
            if localName == "select":
                return khtml.DOM.HTMLSelectElement(node)
            elif localName == "option":
                return khtml.DOM.HTMLOptionElement(node)
            elif localName == "input":
                return khtml.DOM.HTMLInputElement(node)
            elif localName == "button":
                return khtml.DOM.HTMLButtonElement(node)
            elif localName == "form":
                return khtml.DOM.HTMLFormElement(node)
            elif localName == "textarea":
                return khtml.DOM.HTMLTextAreaElement(node)
            else:
                return khtml.DOM.Element(node)

        elif node.nodeType() == xml.dom.Node.ATTRIBUTE_NODE:
            return khtml.DOM.Attr(node)
        elif node.nodeType() == xml.dom.Node.DOCUMENT_NODE:
            return khtml.DOM.Document(node)
        elif node.nodeType() == xml.dom.Node.TEXT_NODE:
            return khtml.DOM.Text(node)
        else:
            return node

    # Methods which could not be generated using the utility methods:
    # Document:

    def importNode(self, id, nodeId, deep):
        if not self.ready:
            return self._returnObject(None, "Node")

        obj, node = map(self._getObject, (id, nodeId))
        return self._returnObject(obj.importNode(node, deep), "Node")

    # NodeList and NamedNodeMap:

    def item(self, id, index):
        if not self.ready:
            return self._returnObject(None, "Node")

        obj = self._getObject(id)
        return self._returnObject(obj.item(index), "Node")

    # Utility methods.

    def string_method(self, method_name):

        """
        Access to the method with the given 'method_name', accepting string
        arguments and returning a string.
        """

        def fn(self, id, *args):
            if not self.ready:
                return self._returnObject(None, "string")
            obj = self._getObject(id)
            args = map(DOMString, args)
            return self._returnObject(getattr(obj, method_name)(*args).string(), "string")
        return new.instancemethod(fn, self, AutomationRoot)

    def other_method(self, method_name):

        """
        Access to the method with the given 'method_name', accepting string
        arguments and returning a boolean or numeric value.
        """

        def fn(self, id, *args):
            if not self.ready:
                return 0
            obj = self._getObject(id)
            args = map(DOMString, args)
            return getattr(obj, method_name)(*args)
        return new.instancemethod(fn, self, AutomationRoot)

    def object_method(self, method_name, result_type):

        """
        Access to the method with the given 'method_name', accepting string
        arguments and returning an object of the given 'result_type'.
        """

        def fn(self, id, *args):
            if not self.ready:
                return self._returnObject(None, result_type)
            obj = self._getObject(id)
            args = map(DOMString, args)
            return self._returnObject(getattr(obj, method_name)(*args), result_type)
        return new.instancemethod(fn, self, AutomationRoot)

    def set_method(self, method_name):

        """
        Access to the method with the given 'method_name', accepting string
        arguments and returning nothing.
        """

        def fn(self, id, *args):
            if not self.ready:
                return
            obj = self._getObject(id)
            args = map(DOMString, args)
            getattr(obj, method_name)(*args)
        return new.instancemethod(fn, self, AutomationRoot)

    def flag_method(self, method_name):

        """
        Access to the method with the given 'method_name', accepting boolean
        or numeric arguments and returning nothing.
        """

        def fn(self, id, *args):
            if not self.ready:
                return
            obj = self._getObject(id)
            getattr(obj, method_name)(*args)
        return new.instancemethod(fn, self, AutomationRoot)

    def action_method(self, method_name, result_type):

        """
        Access to the method with the given 'method_name', accepting node arguments
        and returning an object of the given 'result_type'.
        """

        def fn(self, id, *args):
            if not self.ready:
                return self._returnObject(None, result_type)
            obj = self._getObject(id)
            args = map(self._getObject, args)
            return self._returnObject(getattr(obj, method_name)(*args), result_type)
        return new.instancemethod(fn, self, AutomationRoot)

# vim: tabstop=4 expandtab shiftwidth=4
